(* $Id: miniui_tty.ml 247 2004-01-07 21:42:37Z gerd $
 * ----------------------------------------------------------------------
 * Copyright 2003 by Gerd Stolpmann.
 *
 * This file is part of godi-tools.
 *
 * godi-tools is free software; you can redistribute it and/or modify it
 * under the conditions in the file LICENSE.
 *
 * godi-tools is distributed WITHOUT ANY WARRANTY, as detailed in the
 * file LICENSE.
 *)

open Miniui
open Godi_tty

module Int = struct
  type t = int
  let compare = (compare : int -> int -> int)
end


module IntMap = Map.Make(Int);;


type line =
    { lmargin : int;
      rmargin : int;
      linecnt : string;
    }


let empty_line =
  { lmargin = 0;
    rmargin = 0;
    linecnt = "";
  }


type button =
    { bt_text : string option;
      bt_underline : int option;
      bt_callback : unit -> unit;
      bt_key : key_spec;
    }


type text_entry =
    { te_prompt : string;
      te_callback_ok : string -> unit;
      te_callback_cancel : unit -> unit;
    }


let ctrl_re = Str.regexp "[\000-\031\127-\159]+"

let digit_re = Str.regexp "[0-9]"

let char_re = Str.regexp "[\032-\126\160-\255]"


let this_term = lazy (Godi_tty.get_terminal_capabilities());;


(* backing store: *)
let this_screen = lazy (Array.make (Lazy.force this_term).lines "");;

class tty_button_id : button_id = object end
;;

class tty_page () : page =
object (self)
  val width = (Lazy.force this_term).columns
  val height = (Lazy.force this_term).lines

  val term = Lazy.force this_term
  val screen = Lazy.force this_screen

  (* --- Display the contents of the regions on the screen --- *)

  (* contents of the regions: *)
  val mutable head_rg = IntMap.empty
  val mutable body_rg = IntMap.empty
  val mutable foot_rg = IntMap.empty

  (* highest line numbers + 1: *)
  val mutable head_len = 0
  val mutable body_len = 0
  val mutable foot_len = 0

  (* Scrolling: *)
  val mutable body_start = 0
  val mutable body_auto_scroll = false

  (* Height: *)
  val mutable head_height = 0

  val mutable screen_full_update = true
  val mutable screen_update = true

  method private add_line
                   rg len
                   ?(line = `Append) ?(lmargin = 0) ?(rmargin = 0) s =
    let s' = Str.global_replace ctrl_re "" s in  (* Remove all ctrl chars *)
    let l = { lmargin = lmargin;
	      rmargin = rmargin;
	      linecnt = s' } in
    let n =
      match line with
	  `Append -> len
	| `Replace k -> k in
    let need_update =
      try
	let old_l = IntMap.find n rg in
	old_l <> l
      with
	  Not_found -> true
    in
    if need_update then (
      (* TODO: Setting screen_update here is not optimal. The line might
       * not be visible.
       *)
      screen_update <- true;
      let len' = max len (n+1) in
      let rg' = IntMap.add n l rg in
      (rg', len')
    )
    else
      (rg, len)


  method private put_head_line ?line ?lmargin ?rmargin s =
    let (rg', len') =
      self#add_line head_rg head_len ?line ?lmargin ?rmargin s in
    head_rg <- rg';
    head_len <- len'


  method private put_body_line ?line ?lmargin ?rmargin s =
    let (rg', len') =
      self#add_line body_rg body_len ?line ?lmargin ?rmargin s in
    body_rg <- rg';
    body_len <- len'


  method private put_foot_line ?line ?lmargin ?rmargin s =
    let (rg', len') =
      self#add_line foot_rg foot_len ?line ?lmargin ?rmargin s in
    foot_rg <- rg';
    foot_len <- len'


  method private draw_line row l =
    if row >= 0 && row < height then (
      let n = max 0 (width - l.lmargin - l.rmargin) in

      let s =
	(if l.lmargin > 0 then
	  String.make l.lmargin ' '
	else
	  "")
	^
	(if n > String.length l.linecnt then
	   l.linecnt
	 else
	   String.sub l.linecnt 0 n) in
      if screen_full_update || screen.(row) <> s || row=height-1 then (
	screen.(row) <- s;
	print_string term.to_bol;
	print_string s;
	print_string term.clear_eol;
      )
    )

  method private body_height =
    max 0 (height - head_height - foot_len)


  method private redraw() =
    if screen_update || screen_full_update then (
      print_string (if screen_full_update then term.clear else term.home);
      for i = 0 to head_height - 1 do
	let l = try IntMap.find i head_rg with Not_found -> empty_line in
	self#draw_line i l;
	print_string term.to_bol_next;
      done;
      let body_height = self#body_height in
      for i = body_start to body_start + body_height - 1 do
	let l = try IntMap.find i body_rg with Not_found -> empty_line in
	self#draw_line (i-body_start+head_height) l;
	print_string term.to_bol_next;
      done;
      for i = 0 to foot_len - 1 do
	let l = try IntMap.find i foot_rg with Not_found -> empty_line in
	self#draw_line (head_height + body_height + i) l;
	if i < foot_len - 1 then
	  print_string term.to_bol_next;
      done;
      flush stdout;
      screen_update <- false;
      screen_full_update <- false
    )


  method private at_bottom() =
    let body_max = max 0 (body_len - self#body_height) in
    body_start >= body_max


  method private scroll_to_bottom() =
    body_start <- max 0 (body_len - self#body_height)


  (* --- The foot region: --- *)

  val mutable buttons = []
	        (* Both foot and body buttons. Body buttons have simply
		 * bt_text = None.
		 *)

  val mutable input_prompt = ""
  val mutable input_value = ""

  val mutable busy = false

  method private update_foot_region () =
    foot_rg <- IntMap.empty;
    foot_len <- 0;
    let body_max = max 0 (body_len - self#body_height) in
    let foot_delim =
      if body_start < body_max then
	String.make (max (width-16) 0) '-' ^ "(more)----------"
      else
	String.make width '-' in
    self#put_foot_line foot_delim;
    let line = ref "" in
    let blist = if busy then [] else buttons in
    List.iter
      (fun (id, bt) ->
	 match bt.bt_text with
	     None -> ()
	   | Some text ->
	       let text' =
		 match bt.bt_underline with
		     None -> text
		   | Some k ->
		       String.sub text 0 k ^ "[" ^ String.sub text k 1 ^
		       "]" ^ String.sub text (k+1) (String.length text - k - 1)
	       in
	       let line' =
		 (if !line = "" then "" else !line ^ " ") ^ text' in
	       if String.length line' > width then (
		 self#put_foot_line !line;
		 line := text'
	       )
	       else
		 line := line'
      )
      blist;

    if busy then
      line := "busy...";

    if !line <> "" then self#put_foot_line !line;

    if not busy then begin
      (* CHECK: p > width *)
      let p = String.length input_prompt in
      let v = String.length input_value in
      if p + v >= width then (
	let n = width - p - 2 in
	let k = v - n in
	self#put_foot_line
	  (input_prompt ^ "$" ^ String.sub input_value k n)
      )
      else (
	self#put_foot_line (input_prompt ^ input_value)
      )
    end

  (* --- Input --- *)

  val mutable text_entry = None


  method private init_input() =
    input_prompt <- "> ";
    input_value <- "";
    text_entry <- None;


  method private process_decnum() =
    (* Somebody has entered a decimal number and pressed RETURN *)
    if input_value <> "" then (
      let n = int_of_string input_value in
      let action =
	try
	  ( snd ( List.find
		    (fun (_,bt) -> bt.bt_key = `Number n)
		    buttons
		)
	  ).bt_callback
	with
	    Not_found -> (fun () -> ()) in
      input_value <- "";
      action();
    )


  method private process_te_return() =
    (* Text entry, and the user presses RETURN *)
    match text_entry with
	None ->
	  () (* ??? *)
      | Some te ->
	  let v = input_value in
	  self#init_input();
	  te.te_callback_ok v


  method private process_te_escape() =
    (* Text entry, and the user presses ESC *)
    match text_entry with
	None ->
	  () (* ??? *)
      | Some te ->
	  self#init_input();
	  te.te_callback_cancel()


  method private process_key key =
    let normal_key_re, special_keys =
      match text_entry with
	  None ->
	    (* No activated text entry. The normal keys are 0-9,
	     * and the special keys are taken from the button definitions.
	     * RETURN is also special.
	     *)
	    let skeys =
	      List.flatten
		(List.map
		   (fun (_,bt) ->
		      match bt.bt_key with
			| `Keypress 'p' ->    (* Warning: HACK! *)
			    [ `Char 'p', bt.bt_callback;
			      `PageUp, bt.bt_callback ]
			| `Keypress 'n' ->
			    [ `Char 'n', bt.bt_callback;
			      `PageDown, bt.bt_callback ]
			| `Keypress '<' ->
			    [ `Char '<', bt.bt_callback;
			      `Beginning, bt.bt_callback ]
			| `Keypress '>' ->
			    [ `Char '>', bt.bt_callback;
			      `End, bt.bt_callback ]
			| `Keypress c ->
			    [`Char c, bt.bt_callback]
			| `Number _   ->
			    []
		   )
		   buttons) in

	    let return_keys =
	      [ `Return, self#process_decnum ] in

	    (digit_re, return_keys @ skeys)

	| Some te ->
	    (* All real characters are normal. Special chars are
	     * RETURN and ESCAPE.
	     *)

	    let return_keys =
	      [ `Return, self#process_te_return ] in

	    let escape_key =
	      ( `ESC, self#process_te_escape ) in


	    (char_re, return_keys @ [ escape_key ])
    in

    let key_s =
      match key with
	  `Char c -> String.make 1 c
	| _ -> "" in

    let have_change =
      if Str.string_match normal_key_re key_s 0 then (
	input_value <- input_value ^ key_s;
	true
      )
      else if key = `BS then (
	input_value <> "" && (
	  input_value <- String.sub input_value 0 (String.length input_value-1);
	  true)
      )
      else (
	let key_lc =
	  match key with
	      `Char c -> `Char (Char.lowercase c)
	    | _       -> key in
	let action, ch =
	  try (List.assoc key_lc special_keys, true)
	  with
	      Not_found -> (fun () -> ()), false in
	action();
	ch
      )
    in

    if have_change then
      self#update_foot_region();

    have_change



  method private poll_keypress ?(wait=false) () =
    let key = Godi_tty.poll_keyboard ~wait () in
    key <> `None && self#process_key key


  (* --- Interface --- *)

  val mutable busy_changed = false

  method set_busy b =
    busy_changed <- b <> busy;
    busy <- b

  method busy =
    busy

  method scroll_position =
    self # update_foot_region();
    let max_start = max 0 (body_len - self#body_height) in
    if max_start = 0 then
      1.0
    else
      min 1.0 (float body_start /. float max_start)

  method set_scroll_position p =
    self # update_foot_region();
    let p = max 0.0 (min 1.0 p) in
    let max_start = max 0 (body_len - self#body_height) in
    body_start <- truncate(p *. float max_start)

  method configure ?head_region_height ?auto_scroll () =
    ( match head_region_height with
	  None -> ()
	| Some h -> head_height <- h
    );
    ( match auto_scroll with
	  None -> ()
	| Some b -> body_auto_scroll <- b
    );

  method width() = width

  method height ?(region = `Body) () =
    match region with
	`Head -> head_height
      | `Body -> self#body_height

  method clear () =
    head_rg <- IntMap.empty;
    body_rg <- IntMap.empty;
    foot_rg <- IntMap.empty;
    head_len <- 0;
    body_len <- 0;
    foot_len <- 0;
    body_start <- 0;
    body_auto_scroll <- false;
    head_height <- 0;
    buttons <- [];
    self#add_default_buttons();
    self#init_input();
    self#set_busy false;
    screen_update <- true;
    screen_full_update <- true;

  method put_line ?(region = `Body) ?line ?lmargin ?rmargin s =
    match region with
	`Head -> self#put_head_line ?line ?lmargin ?rmargin s
      | `Body -> let at_bottom = self#at_bottom() in
                 self#put_body_line ?line ?lmargin ?rmargin s;
	         if body_auto_scroll && at_bottom then self#scroll_to_bottom();

  method get_line ?(region = `Body) n =
    let rg =
      match region with
	  `Head -> head_rg
	| `Body -> body_rg in
    try
      (IntMap.find n rg).linecnt
    with
	Not_found -> ""

  method length ?(region = `Body) () =
    match region with
	`Head -> head_len
      | `Body -> body_len

  method add_body_button ~line ?(left=0) ?(width=0) ?(callback = fun _ -> ())
                         key =
    (* line, left, width are ignored by this implementation *)
    let bt =
      { bt_text = None;
	bt_underline = None;
	bt_callback = callback;
	bt_key = key;
      } in

    let id = new tty_button_id in
    buttons <- buttons @ [id, bt];
    id

  method add_foot_button ?text ?underline ?(callback = fun _ -> ())  key =
    let bt =
      { bt_text = text;
	bt_underline = underline;
	bt_callback = callback;
	bt_key = key;
      } in

    let id = new tty_button_id in
    buttons <- buttons @ [id, bt];
    id

  method remove_button id =
    buttons <- List.remove_assoc id buttons

  method activate_text_entry ?(value="") ?(prompt="Input> ")
                             ?(callback_ok = fun _ -> ())
			     ?(callback_cancel = fun _ -> ()) () =
    self # init_input();
    input_prompt <- prompt;
    input_value <- value;
    text_entry <- Some { te_prompt = prompt;
			 te_callback_ok = callback_ok;
			 te_callback_cancel = callback_cancel };
    self # set_busy false

  val mutable processed_refresh_event = 0.0

  method private refresh_event =
    forced_refresh_event() > processed_refresh_event

  method update ?(force=false) () =
    self # update_foot_region();
    if force then screen_full_update <- true;
    let t_refresh = forced_refresh_event() in
    if t_refresh > processed_refresh_event then (
      screen_full_update <- true;
      processed_refresh_event <- t_refresh;
    );
    self # redraw();
    busy_changed <- false;

  method poll_event () =
    self # set_busy false;
    if busy_changed then
      self # update();
    let have_change =
      self # poll_keypress ~wait:false () in
    if have_change || self#refresh_event then
      self # update()

  val mutable quit = false

  method quit() = quit <- true

  method loop() =
    quit <- false;
    self # update();
    while not quit do
      let have_change =
	self # poll_keypress ~wait:true () in
      if have_change || self#refresh_event then
	self # update()
    done;
    quit <- false

  method close() =
    print_string term.clear;
    flush stdout

  method new_page() =
    new tty_page ()


  (* --- Default bindings --- *)

  initializer
    print_string term.init;
    self # init_input();
    self # add_default_buttons();

  method private add_default_buttons() =
    ignore(self#add_foot_button
	     ~text:"prev" ~underline:0 ~callback:self#prev (`Keypress 'p'));
    ignore(self#add_foot_button
	     ~text:"next" ~underline:0 ~callback:self#next (`Keypress 'n'));
    ignore(self#add_foot_button
	     ~callback:self#jump_start (`Keypress '<'));
    ignore(self#add_foot_button
	     ~callback:self#jump_end (`Keypress '>'));
    ignore(self#add_foot_button
	     ~callback:self#full_update (`Keypress '\012'));  (* Ctrl-L *)


  method private full_update() =
    screen_full_update <- true;
    self # update();

  method private prev() =
    let h = self#body_height in
    body_start <- max 0 (body_start - h)

  method private next() =
    let h = self#body_height in
    let m = max 0 (body_len - h) in
    let old_body_start = body_start in
    body_start <- min m (body_start + h);
    if body_start = old_body_start then
      body_auto_scroll <- true


  method private jump_start() =
    body_start <- 0

  method private jump_end() =
    let h = self#body_height in
    let m = max 0 (body_len - h) in
    body_start <- m;
    body_auto_scroll <- true

end
