(* $Id: godi_update.ml 260 2004-01-28 15:53:50Z gerd $
 * ----------------------------------------------------------------------
 * Copyright 2003 by Gerd Stolpmann.
 *
 * This file is part of godi-tools.
 *
 * godi-tools is free software; you can redistribute it and/or modify it
 * under the conditions in the file LICENSE.
 *
 * godi-tools is distributed WITHOUT ANY WARRANTY, as detailed in the
 * file LICENSE.
 *)

open Godi_types
open Godi_exec
open Godi_query
open Godi_versions
open Printf


let available_current = ref None;;


let update_available ?interactivity ?protocol () =
  let makefile = get_tools_makefile() in
  call_cmd
    ?stdout:protocol
    ?stderr:protocol
    ?interactivity
    ~cd:(buildbase())
    ( "godi_make",
      [| "-f"; makefile;
	 "update_available";
      |]
    );
  available_current := None
;;



let words_re = Str.regexp "[ \t]+";;

let parse_available ?protocol () =
  let out s =
    match protocol with
	None -> prerr_string s
      | Some p -> collect p s
  in

  match !available_current with
      Some l -> l
    | None ->
	( let bbase = buildbase() in
	  let avail_cur = Filename.concat bbase "available.current" in
	  let f = open_in avail_cur in
	  let l = ref [] in
	  let n = ref 0 in
	  try
	    while true do
	      incr n;
	      let line = input_line f in
	      let words = Str.split words_re line in
	      match words with
		  [ name; dir; version ] ->
		    l := (name, (dir, version)) :: !l
		| _ ->
		    out (sprintf "Warning: Parsing error in line %d of %s\n"
			   !n avail_cur);
	    done;
	    assert false
	  with
	      End_of_file ->
		close_in f;
		l := List.rev !l;
		available_current := Some !l;
		!l
	    | error ->
		close_in f;
		raise error
	)
;;


let install_build_instructions_for ?interactivity ?protocol
                                   pkgname version pkgrev =

  let makefile = get_tools_makefile() in

  call_cmd
    ?stdout:protocol
    ?stderr:protocol
    ?interactivity
    ( "godi_make",
      [| "-f"; makefile;
	 "install_build_instructions";
	 "pkgname=" ^ pkgname;
	 "version=" ^ version;
	 "pkgrev=" ^ string_of_int pkgrev;
      |]
    )
;;


let update_build_instructions_for ?interactivity ?protocol pkgname =
  let avail_cur = parse_available ?protocol () in
  let (dir,version_pkgrev) =
    try
      List.assoc pkgname avail_cur
    with
	Not_found ->
	  failwith ("No such package: " ^ pkgname) in

  let (_, avail_version, avail_pkgrev) =
    split_name_version ("x-" ^ version_pkgrev) in

  (* Now, if [version] is newer than the installed instance, the
   * build.tgz must be downloaded and extracted.
   *)
  let src_pkgs = scan_source_packages ?protocol () in
  let is_newer =
    try
      let inst_pkg = List.find (fun pkg ->
				  (source_pkg pkg)#name = pkgname) src_pkgs in
      let inst_version = (source_pkg inst_pkg) # version in
      let inst_pkgrev  = (source_pkg inst_pkg) # pkgrevision in

      let n1 = cmp_version_string avail_version inst_version in
      let n2 = if n1=0 then avail_pkgrev - inst_pkgrev else n1 in

      n2 > 0

    with
	Not_found -> true in

  if is_newer then
    install_build_instructions_for
      ?interactivity ?protocol pkgname avail_version avail_pkgrev
;;


let update_build_instructions ?interactivity ?protocol  () =
  let avail_cur = parse_available ?protocol () in
  let pkglist = List.map fst avail_cur in
  List.iter
    (update_build_instructions_for ?interactivity ?protocol)
    pkglist
;;
